from .metrics import get_metric_func, prc_auc, bce, rmse, bounded_mse, bounded_mae, \
    bounded_rmse, accuracy, f1_metric, mcc_metric, sid_metric, wasserstein_metric
from .loss_functions import get_loss_func, bounded_mse_loss, \
    mcc_class_loss, mcc_multiclass_loss, sid_loss, wasserstein_loss
from .cross_validate import mixprop_train, cross_validate, TRAIN_LOGGER_NAME
from .evaluate import evaluate, evaluate_predictions
from .make_predictions import mixprop_predict, make_predictions, load_model
from .molecule_fingerprint import mixprop_fingerprint, model_fingerprint
from .predict import predict
from .run_training import run_training
from .train import train

__all__ = [
    'mixprop_train',
    'cross_validate',
    'TRAIN_LOGGER_NAME',
    'evaluate',
    'evaluate_predictions',
    'mixprop_predict',
    'mixprop_fingerprint',
    'make_predictions',
    'load_model',
    'predict',
    'run_training',
    'train',
    'get_metric_func',
    'prc_auc',
    'bce',
    'rmse',
    'bounded_mse',
    'bounded_mae',
    'bounded_rmse',
    'accuracy',
    'f1_metric',
    'mcc_metric',
    'sid_metric',
    'wasserstein_metric',
    'get_loss_func',
    'bounded_mse_loss',
    'mcc_class_loss',
    'mcc_multiclass_loss',
    'sid_loss',
    'wasserstein_loss'
]



import numpy as np
import pandas as pd


def report_stats(nist_knovel_all, test_mols):
    """
    Prints the number of molecules/datapoints in the test and training sets.
    
    nist_knovel_all: pandas dataframe containing the dataset
    test_mols: list containing compounds to held out of the training set and assigned to the test set    
    """

    # Average over all data that looks like duplicates before checkpointing
    nist_knovel_all_nodup = nist_knovel_all.groupby(
        ["MOL_1", "MOL_2", "MolFrac_1"]
    ).mean()
    nist_knovel_all_nodup.reset_index(inplace=True)

    nist_knovel_all_nodup["test_1"] = nist_knovel_all_nodup["MOL_1"].apply(
        lambda smi: smi in test_mols
    )
    nist_knovel_all_nodup["test_2"] = nist_knovel_all_nodup["MOL_2"].apply(
        lambda smi: smi in test_mols
    )
    nist_knovel_all_nodup = nist_knovel_all_nodup[
        ["MOL_1", "MOL_2", "MolFrac_1", "logV", "test_1", "test_2"]
    ]
    test_data = nist_knovel_all_nodup[
        nist_knovel_all_nodup["test_1"] | nist_knovel_all_nodup["test_2"]
    ].dropna()
    train_data = nist_knovel_all_nodup[
        ~(nist_knovel_all_nodup["test_1"] | nist_knovel_all_nodup["test_2"])
    ].dropna()

    print(
        "Total Number of Molecules:{}".format(
            len(
                list(
                    set(
                        list(nist_knovel_all_nodup["MOL_1"].values)
                        + list(nist_knovel_all_nodup["MOL_2"].values)
                    )
                )
            )
        )
    )
    print(
        "Number of Molecules in Training Set:{}".format(
            len(
                list(
                    set(
                        list(train_data["MOL_1"].values)
                        + list(train_data["MOL_2"].values)
                    )
                )
            )
        )
    )
    print(
        "Number of Molecules in Test Set:{}".format(
            len(
                list(
                    set(
                        list(test_data["MOL_1"].values)
                        + list(test_data["MOL_2"].values)
                    )
                )
            )
        )
    )
    print("Number of Datapoints in Training Set:{}".format(len(train_data)))
    print("Number of Datapoints in Test Set:{}".format(len(test_data)))
    
    return nist_knovel_all_nodup


def series_std(data):
    """
    Obtains the standard deviation between multiple series contained in a dictionary.
    
    data: dictionary containing one or more series
    """

    flag = 0
    std_list = []
    for x_val, y_val in data.items():
        if len(y_val) > 1:
            flag = 1
            std_list.append(np.std(np.array(y_val)))
    if flag == 1:
        return np.mean(std_list)


def assign_phase(T, smi1, smi2, df_mols, mp_buffer=10, bp_buffer=0):
    bp1 = df_mols[df_mols["SMILES"] == smi1]["BP"].values[0]
    bp2 = df_mols[df_mols["SMILES"] == smi2]["BP"].values[0]
    mp1 = df_mols[df_mols["SMILES"] == smi1]["MP"].values[0]
    mp2 = df_mols[df_mols["SMILES"] == smi2]["MP"].values[0]
    if mp1 > (T - mp_buffer):
        phase1 = "solid"
    elif (mp1 <= (T - mp_buffer)) & (bp1 > (T + bp_buffer)):
        phase1 = "liquid"
    else:
        phase1 = "gas"

    if mp2 > (T - mp_buffer):
        phase2 = "solid"
    elif (mp2 <= (T - mp_buffer)) & (bp2 > (T + bp_buffer)):
        phase2 = "liquid"
    else:
        phase2 = "gas"
    return phase1, phase2


def soft_assign_phase(T, smi1, smi2, nist_knovel_all_mols_df):
    try:
        return assign_phase(T, smi1, smi2, nist_knovel_all_mols_df)
    except:
        return None, None



import numpy as np
import pandas as pd


def process_nist(nist):
    """
    Apply standard processing for NIST dataset.
    Note that these settings are specific to the dataset provided.
    """

    nist["logV"] = np.log10(nist["Visc"] * 1000)

    nist = nist.drop_duplicates()
    nist = nist[nist["MOL_1"] != "MISSING_ID"]
    nist = nist[nist["MOL_2"] != "MISSING_ID"]

    return nist


def process_dippr(knovel, nist, args):
    """
    Apply standard processing for DIPPR dataset.
    Note that these settings are specific to the dataset provided.
    """

    nist_all_mols = list(set(list(nist["MOL_1"].values) + list(nist["MOL_2"].values)))

    knovel = knovel[["SMILES", "logV"]].dropna().drop_duplicates()
    knovel["in_nist"] = knovel["SMILES"].apply(lambda smi: smi in nist_all_mols)
    most_common_mols = [args["dummy_mol"]]
    knovel_mix = pd.DataFrame()

    for smi, logV, in_nist in knovel.values:
        if not in_nist:
            for dummy in most_common_mols:
                add_dict1 = {
                    "MOL_1": smi,
                    "MOL_2": dummy,
                    "MolFrac_1": 1.0,
                    "logV": logV,
                }
                add_dict2 = {
                    "MOL_1": dummy,
                    "MOL_2": smi,
                    "MolFrac_1": 0.0,
                    "logV": logV,
                }
                add_df = pd.DataFrame([add_dict1, add_dict2])
            knovel_mix = pd.concat((knovel_mix, add_df))

    knovel_mix["T"] = 298
    knovel_mix["Ref_ID"] = "ref1"
    return knovel_mix


def load_data(args):
    """
    Load and process NIST and DIPPR datasets.
    """

    nist = pd.read_csv(args["NIST"])
    dippr = pd.read_csv(args["DIPPR"])
    nist = process_nist(nist)
    dippr = process_dippr(dippr, nist, args)
    nist_knovel_all = pd.concat((nist, dippr))

    return nist_knovel_all



import numpy as np
import os
import pandas as pd
from pathlib import Path

from mixprop.train import predict
from mixprop.data import MoleculeDataset, MoleculeDataLoader, MoleculeDatapoint
from mixprop.utils import load_args, load_checkpoint, load_scalers

from rdkit import Chem


class mixprop_model():
    
    def __init__(self, checkpoint_dir):
        self.checkpoints = []
        for root, _, files in os.walk(checkpoint_dir):
            for fname in files:
                if fname.endswith('.pt'):
                    fname = os.path.join(root, fname)
                    scalers =load_scalers(fname)
                    self.scaler, self.features_scaler = scalers[0], scalers[1]
                    self.train_args = load_args(fname)
                    model = load_checkpoint(fname) #, cuda=True)
                    self.checkpoints.append(model)

    def __call__(self, args):
                
        if args['n_models']==None:
            args['n_models']=len(self.checkpoints)
        assert args['n_models']<=len(self.checkpoints),'Too many models requested. {} models requested.'.format(args['n_models'])
        assert args['n_models']>1, 'Multiple models are needed for reliability analysis.'
        
        model_input= MoleculeDatapoint(smiles=[args['smi1'],args['smi2']],features=[args['molfrac1'],args['T']])
        model_input = MoleculeDataset([model_input])
        model_input_loader = MoleculeDataLoader(dataset=model_input,num_workers=args['num_workers'])

        all_model_preds = []
        for model in self.checkpoints[:args['n_models']]:
             model_preds = predict(
                            model=model,
                            data_loader=model_input_loader,
                            scaler=self.scaler,
                            disable_progress_bar=True)
             all_model_preds.append(model_preds)
        
        avg_prediction = np.mean(all_model_preds)
        reliability = np.var(all_model_preds)<args['threshold']
        return avg_prediction,reliability

def load_model(args):
    if 'checkpoint_dir' in args.keys():
        return mixprop_model(args['checkpoint_dir'])
    else:
        path = str(Path(__file__).absolute())
        path = '/'.join(path.split('/')[:-1])
        checkpoint_dir = os.path.join(path,'pretrained_models/nist_dippr_model/nist_dippr_model')
        
        print('Loading models from {}'.format(checkpoint_dir))
        
        return mixprop_model(checkpoint_dir)



def onepoint_assertions(args):
    
    # Requirements:
    assert type(args['smi1'])==str, 'Molecules need to be input as SMILES strings.'
    assert type(args['smi2'])==str, 'Molecules need to be input as SMILES strings.'
    assert (type(args['T'])==float)|(type(args['T'])==int)|(type(args['T'])==np.int64)|(type(args['T'])==np.float), 'Temperature needs to be input as a number.'
    assert (type(args['molfrac1'])==float)|(type(args['molfrac1'])==int)|(type(args['T'])==np.int64)|(type(args['T'])==np.float), 'Mole fraction needs to be input as a number.'
    
    assert (args['molfrac1']>=0.0)&(args['molfrac1']<=1.0), 'Mole fraction needs to be between 0 and 1.'
    
    mol1 = Chem.MolFromSmiles(args['smi1'])
    mol2 = Chem.MolFromSmiles(args['smi2'])
    assert type(mol1)==Chem.rdchem.Mol, 'Invalid SMILES entry, please check that your SMILES are valid.'
    assert type(mol2)==Chem.rdchem.Mol, 'Invalid SMILES entry, please check that your SMILES are valid.'

    
    # Warnings:
    if ('.' in args['smi1'])|('.' in args['smi2']):
        print('\nWARNING: Multiple molecules are contained within a single SMILES. Predictions may be unreliable.')
    
    if (args['T']<293)|(args['T']>323):
        print('\nWARNING: Temperature is outside of recommended range (293 < T < 323). Predictions may be unreliable.')


def visc_pred_onepoint(model, args):
          
    # Check input validity:
    onepoint_assertions(args)
    
    prediction_log,reliability = model(args)
    prediction = 10**prediction_log #Prediction must be converted cP units (without the log)
    
    return prediction, reliability


def visc_pred_single(args):
    
    
    model = load_model(args)
    prediction, reliability = visc_pred_onepoint(model, args)
    
    return prediction, reliability
    

def visc_pred_T_curve(args):
    
    model = load_model(args)
    
    T_low = 293
    T_high = 323
    interval = 5
    T_vals = np.arange(T_low,T_high+interval,interval)
    
    preds, rels = [], []
    for T in T_vals:
        T = float(T)
        args['T'] = T
        prediction, reliability = visc_pred_onepoint(model, args)
        preds.append(prediction)
        rels.append(reliability)

    return preds, T_vals, rels


def visc_pred_molfrac1_curve(args):
    
    model = load_model(args)
    
        
    frac_low = 0.0
    frac_high = 1.0
    interval = 0.1
    frac_vals = np.arange(frac_low,frac_high+interval,interval)
    
    preds, rels = [], []
    for frac in frac_vals:
        frac = float(frac)
        args['molfrac1'] = frac
        prediction, reliability = visc_pred_onepoint(model, args)
        preds.append(prediction)
        rels.append(reliability)

    return preds, frac_vals, rels


def visc_pred_read_csv(args): # Need additional path argument
    
    model = load_model(args)
    
    data = pd.read_csv(args['input_path'])
    cols = data.columns
    
    preds, rels = [], []
    for i in range(len(data)):
        args['smi1'] = data[cols[0]].iloc[i]
        args['smi2'] = data[cols[1]].iloc[i]
        args['molfrac1'] = data[cols[2]].iloc[i]
        args['T'] = data[cols[3]].iloc[i]

        prediction, reliability = visc_pred_onepoint(model, args)
        preds.append(prediction)
        rels.append(reliability)
        
    data['Viscoisty Predictions'] = preds
    data['Reliability'] = rels
    
    return preds, rels, data


import numpy as np
import pandas as pd
from .utils import report_stats


def remove_inconsistent(nist_knovel_all, test_mols, args):
    """
    Remove inconsistent data, defined as data where experimental values of viscosity
    vary widely.
    """
    ## Remove based on inconsistencies in pure data:--------------------

    # Create pure dataset (for further cleaning):
    df_1 = pd.DataFrame()
    df_1 = nist_knovel_all[nist_knovel_all["MolFrac_1"] == 1][
        ["MOL_1", "logV", "T"]
    ].rename(columns={"MOL_1": "SMILES"})

    df_2 = pd.DataFrame()
    df_2 = nist_knovel_all[nist_knovel_all["MolFrac_1"] == 0][
        ["MOL_2", "logV", "T"]
    ].rename(columns={"MOL_2": "SMILES"})

    nist_pure = pd.concat((df_1, df_2))
    nist_pure = nist_pure.drop_duplicates()

    nist_pure_mols = pd.DataFrame()
    nist_pure_mols["SMILES"] = [
        x[0] for x in list(nist_pure.groupby(["SMILES", "T"]).mean().index)
    ]
    nist_pure_mols["T"] = [
        x[1] for x in list(nist_pure.groupby(["SMILES", "T"]).mean().index)
    ]
    nist_pure_mols["logV_avg"] = (
        nist_pure.groupby(["SMILES", "T"]).mean()["logV"].values
    )
    nist_pure_mols["logV_std"] = nist_pure.groupby(["SMILES", "T"]).std()["logV"].values

    # Remove molecules that have a pure component STD of greater than 1:
    smi2remove = nist_pure_mols[nist_pure_mols["logV_std"] > 1.0]["SMILES"].values

    nist_knovel_all = nist_knovel_all[
        nist_knovel_all["MOL_1"].apply(lambda smi: smi not in smi2remove)
    ]
    nist_knovel_all = nist_knovel_all[
        nist_knovel_all["MOL_2"].apply(lambda smi: smi not in smi2remove)
    ]

    ## Remove based on inconsistencies in nearly-pure data:---------------

    thresh_pure = args["thresh_pure"]  # Distance from pure considered "pure-ish"
    thresh_logV = args["thresh_logV"]
    nist_near_pure = nist_knovel_all[
        (nist_knovel_all["MolFrac_1"] < thresh_pure)
        | (nist_knovel_all["MolFrac_1"] > (1 - thresh_pure))
    ]

    target_mol_list = []
    target_frac_list = []
    target_id_list = []
    for i in range(len(nist_near_pure)):
        if nist_near_pure.iloc[i]["MolFrac_1"] < 0.5:
            target_mol = nist_near_pure.iloc[i]["MOL_2"]
            target_frac = 1 - nist_near_pure.iloc[i]["MolFrac_1"]
            target_id = nist_near_pure.iloc[i]["ID_2"]
        elif nist_near_pure.iloc[i]["MolFrac_1"] > 0.5:
            target_mol = nist_near_pure.iloc[i]["MOL_1"]
            target_frac = nist_near_pure.iloc[i]["MolFrac_1"]
            target_id = nist_near_pure.iloc[i]["ID_1"]
        target_mol_list.append(target_mol)
        target_frac_list.append(target_frac)
        target_id_list.append(target_id)

    nist_near_pure["TargetMol"] = target_mol_list
    nist_near_pure["TargetFrac"] = target_frac_list
    nist_near_pure["TargetID"] = target_id_list

    nist_near_pure = nist_near_pure[
        [
            "TargetMol",
            "TargetFrac",
            "TargetID",
            "Visc",
            "Visc_Unc",
            "logV",
            "T",
            "P",
            "Ref_ID",
        ]
    ]
    nist_near_pure_medians = pd.DataFrame()

    nist_near_pure_medians["TargetMol"] = [
        x[0] for x in nist_near_pure.groupby(["TargetMol", "T"]).median()["logV"].index
    ]
    nist_near_pure_medians["T"] = [
        x[1] for x in nist_near_pure.groupby(["TargetMol", "T"]).median()["logV"].index
    ]
    nist_near_pure_medians["logV_median"] = (
        nist_near_pure.groupby(["TargetMol", "T"]).median()["logV"].values
    )

    suspicious_index_list = []
    for target_mol, target_median, T in nist_near_pure_medians[
        ["TargetMol", "logV_median", "T"]
    ].values:
        suspicious_index_list.extend(
            nist_near_pure[
                (nist_near_pure["TargetMol"] == target_mol)
                & (nist_near_pure["T"] == T)
                & (np.abs(nist_near_pure["logV"] - target_median) > thresh_logV)
            ].index
        )

    suspicious_data = nist_near_pure.loc[suspicious_index_list]

    nist_knovel_all["suspicious"] = nist_knovel_all["Ref_ID"].apply(
        lambda x: x in suspicious_data["Ref_ID"].values
    )

    nist_knovel_all = nist_knovel_all[~nist_knovel_all["suspicious"]]

    report_stats(nist_knovel_all, test_mols)

    return nist_knovel_all



import numpy as np
import pandas as pd
import os


def write_data(nist_knovel_all, test_mols, input_args):

    # Average over all data that looks like duplicates before checkpointing
    nist_knovel_all_nodup = nist_knovel_all.groupby(
        ["MOL_1", "MOL_2", "MolFrac_1"]
    ).mean()
    nist_knovel_all_nodup.reset_index(inplace=True)

    nist_knovel_all_nodup["test_1"] = nist_knovel_all_nodup["MOL_1"].apply(
        lambda smi: smi in test_mols
    )
    nist_knovel_all_nodup["test_2"] = nist_knovel_all_nodup["MOL_2"].apply(
        lambda smi: smi in test_mols
    )
    nist_knovel_all_nodup = nist_knovel_all_nodup[
        ["MOL_1", "MOL_2", "MolFrac_1", "T", "logV", "test_1", "test_2"]
    ]
    test_data = nist_knovel_all_nodup[
        nist_knovel_all_nodup["test_1"] | nist_knovel_all_nodup["test_2"]
    ].dropna()
    train_data = nist_knovel_all_nodup[
        ~(nist_knovel_all_nodup["test_1"] | nist_knovel_all_nodup["test_2"])
    ].dropna()

    train_data[["MOL_1", "MOL_2", "logV"]].to_csv(
        os.path.join(input_args["out_path"], "data.csv"), index=False
    )
    train_data[["MolFrac_1", "T"]].to_csv(
        os.path.join(input_args["out_path"], "data_features.csv"), index=False
    )
    test_data[["MOL_1", "MOL_2", "logV"]].to_csv(
        os.path.join(input_args["out_path"], "test.csv"), index=False
    )
    test_data[["MolFrac_1", "T"]].to_csv(
        os.path.join(input_args["out_path"], "test_features.csv"), index=False
    )



from .features_generators import get_available_features_generators, get_features_generator, \
    morgan_binary_features_generator, morgan_counts_features_generator, rdkit_2d_features_generator, \
    rdkit_2d_normalized_features_generator, register_features_generator
from .featurization import atom_features, bond_features, BatchMolGraph, get_atom_fdim, get_bond_fdim, mol2graph, \
    MolGraph, onek_encoding_unk, set_extra_atom_fdim, set_extra_bond_fdim, set_reaction, set_explicit_h, \
    set_adding_hs, is_reaction, is_explicit_h, is_adding_hs, is_mol, reset_featurization_parameters
from .utils import load_features, save_features, load_valid_atom_or_bond_features

__all__ = [
    'get_available_features_generators',
    'get_features_generator',
    'morgan_binary_features_generator',
    'morgan_counts_features_generator',
    'rdkit_2d_features_generator',
    'rdkit_2d_normalized_features_generator',
    'atom_features',
    'bond_features',
    'BatchMolGraph',
    'get_atom_fdim',
    'set_extra_atom_fdim',
    'get_bond_fdim',
    'set_extra_bond_fdim',
    'set_explicit_h',
    'set_adding_hs',
    'set_reaction',
    'is_reaction',
    'is_explicit_h',
    'is_adding_hs',
    'is_mol',
    'mol2graph',
    'MolGraph',
    'onek_encoding_unk',
    'load_features',
    'save_features',
    'load_valid_atom_or_bond_features',
    'reset_featurization_parameters'
]



import numpy as np
import pandas as pd
from .utils import soft_assign_phase, report_stats


def remove_not_liquid(nist_knovel_all, test_mols, args):
    """
    Remove compounds that are not liquid at the temperature at which they are reported.
    """

    nist_knovel_all_mols = list(
        set(
            list(nist_knovel_all["MOL_1"].values)
            + list(nist_knovel_all["MOL_2"].values)
        )
    )
    nist_knovel_all_mols_df = pd.DataFrame()
    nist_knovel_all_mols_df["SMILES"] = nist_knovel_all_mols

    # After running chemprop models:
    bp_df = pd.read_csv(args["bp_pred_path"])
    mp_df = pd.read_csv(args["mp_pred_path"])
    bp_df = bp_df[bp_df["BP"] != "Invalid SMILES"]
    mp_df = mp_df[mp_df["MP"] != "Invalid SMILES"]
    nist_knovel_all_mols_df = mp_df
    nist_knovel_all_mols_df["BP"] = bp_df["BP"].apply(lambda x: float(x) + 273)
    nist_knovel_all_mols_df["MP"] = nist_knovel_all_mols_df["MP"].apply(
        lambda x: float(x)
    )

    nist_knovel_all["Phase"] = nist_knovel_all.apply(
        lambda x: soft_assign_phase(
            x["T"], x["MOL_1"], x["MOL_2"], nist_knovel_all_mols_df
        ),
        axis=1,
    )

    nist_knovel_all = nist_knovel_all[nist_knovel_all["Phase"] == ("liquid", "liquid")]

    report_stats(nist_knovel_all, test_mols)

    return nist_knovel_all



from typing import List

import torch
from tqdm import tqdm

from mixprop.data import MoleculeDataLoader, MoleculeDataset, StandardScaler
from mixprop.models import MoleculeModel


def predict(model: MoleculeModel,
            data_loader: MoleculeDataLoader,
            disable_progress_bar: bool = False,
            scaler: StandardScaler = None) -> List[List[float]]:
    """
    Makes predictions on a dataset using an ensemble of models.

    :param model: A :class:`~mixprop.models.model.MoleculeModel`.
    :param data_loader: A :class:`~mixprop.data.data.MoleculeDataLoader`.
    :param disable_progress_bar: Whether to disable the progress bar.
    :param scaler: A :class:`~mixprop.features.scaler.StandardScaler` object fit on the training targets.
    :return: A list of lists of predictions. The outer list is molecules while the inner list is tasks.d
    """
    model.eval()

    preds = []

    for batch in tqdm(data_loader, disable=disable_progress_bar, leave=False):
        # Prepare batch
        batch: MoleculeDataset
        mol_batch, features_batch, atom_descriptors_batch, atom_features_batch, bond_features_batch = \
            batch.batch_graph(), batch.features(), batch.atom_descriptors(), batch.atom_features(), batch.bond_features()


        # Make predictions
        with torch.no_grad():
            batch_preds = model(mol_batch, features_batch, atom_descriptors_batch,
                                atom_features_batch, bond_features_batch)

        batch_preds = batch_preds.data.cpu().numpy()

        # Inverse scale if regression
        if scaler is not None:
            batch_preds = scaler.inverse_transform(batch_preds)

        # Collect vectors
        batch_preds = batch_preds.tolist()
        preds.extend(batch_preds)

    return preds



import csv
import pickle

import numpy as np
from tqdm import tqdm

from mixprop.args import SklearnPredictArgs, SklearnTrainArgs
from mixprop.data import get_data
from mixprop.features import get_features_generator
from mixprop.sklearn_train import predict
from mixprop.utils import makedirs, timeit


@timeit()
def predict_sklearn(args: SklearnPredictArgs) -> None:
    """
    Loads data and a trained scikit-learn model and uses the model to make predictions on the data.

   :param args: A :class:`~mixprop.args.SklearnPredictArgs` object containing arguments for
                 loading data, loading a trained scikit-learn model, and making predictions with the model.
    """
    print('Loading data')
    data = get_data(path=args.test_path,
                    smiles_columns=args.smiles_columns,
                    target_columns=[],
                    ignore_columns=[],
                    store_row=True)

    print('Loading training arguments')
    with open(args.checkpoint_paths[0], 'rb') as f:
        model = pickle.load(f)
        train_args: SklearnTrainArgs = SklearnTrainArgs().from_dict(model.train_args, skip_unsettable=True)

    print('Computing morgan fingerprints')
    morgan_fingerprint = get_features_generator('morgan')
    for datapoint in tqdm(data, total=len(data)):
        for s in datapoint.smiles:
            datapoint.extend_features(morgan_fingerprint(mol=s, radius=train_args.radius, num_bits=train_args.num_bits))

    print(f'Predicting with an ensemble of {len(args.checkpoint_paths)} models')
    sum_preds = np.zeros((len(data), train_args.num_tasks))

    for checkpoint_path in tqdm(args.checkpoint_paths, total=len(args.checkpoint_paths)):
        with open(checkpoint_path, 'rb') as f:
            model = pickle.load(f)

        model_preds = predict(
            model=model,
            model_type=train_args.model_type,
            dataset_type=train_args.dataset_type,
            features=data.features()
        )
        sum_preds += np.array(model_preds)

    # Ensemble predictions
    avg_preds = sum_preds / len(args.checkpoint_paths)
    avg_preds = avg_preds.tolist()

    print(f'Saving predictions to {args.preds_path}')
    # assert len(data) == len(avg_preds)    #TODO: address with unit test later
    makedirs(args.preds_path, isfile=True)

    # Copy predictions over to data
    for datapoint, preds in zip(data, avg_preds):
        for pred_name, pred in zip(train_args.task_names, preds):
            datapoint.row[pred_name] = pred

    # Save
    with open(args.preds_path, 'w') as f:
        writer = csv.DictWriter(f, fieldnames=data[0].row.keys())
        writer.writeheader()

        for datapoint in data:
            writer.writerow(datapoint.row)


def sklearn_predict() -> None:
    """Parses scikit-learn predicting arguments and runs prediction using a trained scikit-learn model.

    This is the entry point for the command line command :code:`sklearn_predict`.
    """
    predict_sklearn(args=SklearnPredictArgs().parse_args())



import mixprop.data
import mixprop.features
import mixprop.models
import mixprop.train

import mixprop.args
import mixprop.constants
import mixprop.hyperparameter_optimization
import mixprop.interpret
import mixprop.nn_utils
import mixprop.utils
import mixprop.rdkit
import mixprop.sklearn_predict
import mixprop.sklearn_train
import mixprop.spectra_utils
import mixprop.hyperopt_utils

from mixprop._version import __version__
